local AddonName, AddonTable = ...

FlapAway = LibStub("AceAddon-3.0"):NewAddon(AddonName, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")

local FA = FlapAway

local L = LibStub("AceLocale-3.0"):GetLocale(AddonName)
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

FA.DebugLevel = 0
-- FA.DebugLevel = 1
-- use 1 for non-spammy stuff and current testing, e.g. response to button click
-- use 2 for common outputs, e.g. interface testing
-- use 3 for very frequent outputs, e.g. once per second
-- use 4 for very spammy stuff like aura update events

-- Configure NPC list
FA.npcInfoTimerCount = 0
FA.npcInfoFoundAll = false
--
FA.npcInfo = {
    [170721] = {npcName = nil, questID = nil, },        -- Instructor Malo
}

FA.questName = {}

FA.InstructorNPCID = 170721

function FA.Debug(DebugLevel, ...)
    if DebugLevel <= FA.DebugLevel then
        FA:Print(...)
    end
end

function FA:CreateFlapAwayFrame(FrameName, FrameTitleText, DefaultPoint, DefaultWidth, DefaultHeight, CloseFuction)

    local Frame

    -- Frame Setup
    Frame = CreateFrame("Frame", FrameName, UIParent)
    Frame:Hide()

    Frame.DefaultPoint = DefaultPoint
    Frame.DefaultWidth = DefaultWidth
    Frame.DefaultHeight = DefaultHeight

    function Frame:Reset()
        Frame:ClearAllPoints()
        Frame:SetPoint(unpack(Frame.DefaultPoint))
        Frame:SetWidth(DefaultWidth)
        Frame:SetHeight(DefaultHeight)
        Frame:SetFrameStrata("DIALOG")
        Frame:SetMovable(true)
        Frame:EnableMouse(true)
        Frame:SetClampedToScreen(true)
    end

    Frame:Reset()

    -- movement handlers are set on TitleButton below

    Frame.TitleBG = Frame:CreateTexture(FrameName .. "TitleBG", "BACKGROUND")
    Frame.TitleBG:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Title-Background")
    Frame.TitleBG:SetPoint("TOPLEFT", 9, -6)
    Frame.TitleBG:SetPoint("BOTTOMRIGHT", Frame, "TOPRIGHT", -28, -24)

    Frame.DialogBG = Frame:CreateTexture(FrameName .. "DialogBG", "BACKGROUND")
    Frame.DialogBG:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
    Frame.DialogBG:SetPoint("TOPLEFT", 8, -24)
    Frame.DialogBG:SetPoint("BOTTOMRIGHT", -6, 8)
    Frame.DialogBG:SetColorTexture(0, 0, 0, .75)

    Frame.TopLeft = Frame:CreateTexture(FrameName .. "TopLeft", "BORDER")
    Frame.TopLeft:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.TopLeft:SetPoint("TOPLEFT")
    Frame.TopLeft:SetWidth(64)
    Frame.TopLeft:SetHeight(64)
    Frame.TopLeft:SetTexCoord(0.501953125, 0.625, 0, 1)

    Frame.TopRight = Frame:CreateTexture(FrameName .. "TopRight", "BORDER")
    Frame.TopRight:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.TopRight:SetPoint("TOPRIGHT")
    Frame.TopRight:SetWidth(64)
    Frame.TopRight:SetHeight(64)
    Frame.TopRight:SetTexCoord(0.625, 0.75, 0, 1)

    Frame.Top = Frame:CreateTexture(FrameName .. "Top", "BORDER")
    Frame.Top:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.Top:SetPoint("TOPLEFT", Frame.TopLeft, "TOPRIGHT")
    Frame.Top:SetPoint("TOPRIGHT", Frame.TopRight, "TOPLEFT")
    Frame.Top:SetWidth(0)
    Frame.Top:SetHeight(64)
    Frame.Top:SetTexCoord(0.25, 0.369140625, 0, 1)

    Frame.BottomLeft = Frame:CreateTexture(FrameName .. "BottomLeft", "BORDER")
    Frame.BottomLeft:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.BottomLeft:SetPoint("BOTTOMLEFT")
    Frame.BottomLeft:SetWidth(64)
    Frame.BottomLeft:SetHeight(64)
    Frame.BottomLeft:SetTexCoord(0.751953125, 0.875, 0, 1)

    Frame.BottomRight = Frame:CreateTexture(FrameName .. "BottomRight", "BORDER")
    Frame.BottomRight:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.BottomRight:SetPoint("BOTTOMRIGHT")
    Frame.BottomRight:SetWidth(64)
    Frame.BottomRight:SetHeight(64)
    Frame.BottomRight:SetTexCoord(0.875, 1, 0, 1)

    Frame.Bottom = Frame:CreateTexture(FrameName .. "Bottom", "BORDER")
    Frame.Bottom:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.Bottom:SetPoint("BOTTOMLEFT", Frame.BottomLeft, "BOTTOMRIGHT")
    Frame.Bottom:SetPoint("BOTTOMRIGHT", Frame.BottomRight, "BOTTOMLEFT")
    Frame.Bottom:SetWidth(0)
    Frame.Bottom:SetHeight(64)
    Frame.Bottom:SetTexCoord(0.376953125, 0.498046875, 0, 1)

    Frame.Left = Frame:CreateTexture(FrameName .. "Left", "BORDER")
    Frame.Left:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.Left:SetPoint("TOPLEFT", Frame.TopLeft, "BOTTOMLEFT")
    Frame.Left:SetPoint("BOTTOMLEFT", Frame.BottomLeft, "TOPLEFT")
    Frame.Left:SetWidth(64)
    Frame.Left:SetHeight(0)
    Frame.Left:SetTexCoord(0.001953125, 0.125, 0, 1)

    Frame.Right = Frame:CreateTexture(FrameName .. "Bottom", "BORDER")
    Frame.Right:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-Border")
    Frame.Right:SetPoint("TOPRIGHT", Frame.TopRight, "BOTTOMRIGHT")
    Frame.Right:SetPoint("BOTTOMRIGHT", Frame.BottomRight, "TOPRIGHT")
    Frame.Right:SetWidth(64)
    Frame.Right:SetHeight(0)
    Frame.Right:SetTexCoord(0.1171875, 0.2421875, 0, 1)

    Frame.CloseButton = CreateFrame("Button", FrameName .. "CloseButton", Frame, "UIPanelCloseButton")
    Frame.CloseButton:SetPoint("TOPRIGHT", 2, 1)
    Frame.CloseButton:SetWidth(32)
    Frame.CloseButton:SetHeight(32)
    Frame.CloseButton:SetScript("OnClick", CloseFuction)

    Frame.TitleButton = CreateFrame("Button", FrameName .. "TitleButton", Frame)
    Frame.TitleButton:SetPoint("TOPLEFT", Frame.TitleBG, "TOPLEFT")
    Frame.TitleButton:SetPoint("BOTTOMRIGHT", Frame.TitleBG, "BOTTOMRIGHT")
    Frame.TitleButton:RegisterForDrag("LeftButton")
    Frame.TitleButton:SetScript("OnDragStart", function(self) local parent = self:GetParent() parent.moving = true parent:StartMoving() end)
    Frame.TitleButton:SetScript("OnDragStop", function(self) local parent = self:GetParent() parent.moving = nil parent:StopMovingOrSizing() end)
    
    Frame.TitleButtonText = Frame.TitleButton:CreateFontString(FrameName .. "TitleButtonText", "OVERLAY", "GameFontNormal")
    Frame.TitleButtonText:SetPoint("TOPLEFT", Frame.TitleButton, "TOPLEFT")
    Frame.TitleButtonText:SetPoint("BOTTOMRIGHT", Frame.TitleButton, "BOTTOMRIGHT")
    Frame.TitleButtonText:SetText(FrameTitleText)

    return Frame
end

function FA:OnInitialize()
    FA:RegisterChatCommand("flapaway", "SlashHandler")

    -- Addon Options
    FA.defaultOptions = {
        global = {
            KEYBIND1 = "MOUSEWHEELDOWN",
			KEYBIND2 = "MOUSEWHEELUP",
        }
    }
    FA.db = LibStub("AceDB-3.0"):New("FlapAwaySettingsDB", FA.defaultOptions)

    -- Create Frames here, so that Layout-Cache will be applied to them
    FA.FAFrame = FA:CreateFlapAwayFrame("FlapAwayFrame", "FlapAway", {"TOP", "UIParent", "TOP", 0, -225}, 320, 100, FA.Close)
	
	FA.FAFrame.Button1 = CreateFrame("Button", "FlapAwayButton1", FA.FAFrame, "InsecureActionButtonTemplate")
	FA.FAFrame.Button1:SetAttribute("type1", "pet")
	FA.FAFrame.Button1:SetAttribute("action", 1)
	
	FA.FAFrame.Button2 = CreateFrame("Button", "FlapAwayButton2", FA.FAFrame, "InsecureActionButtonTemplate")
	FA.FAFrame.Button2:SetAttribute("type1", "pet")
	FA.FAFrame.Button2:SetAttribute("action", 1)

end

function FA:OnEnable()
    FA.db.RegisterCallback(FA, "OnProfileChanged", "RefreshConfig")
    FA.db.RegisterCallback(FA, "OnProfileCopied", "RefreshConfig")
    FA.db.RegisterCallback(FA, "OnProfileReset", "RefreshConfig")
	
    local options = {
        type = "group",
        args = {
            generalgroup = {
                type = "group",
                name = L["General Settings Group"],
                inline = true,
                order = 10,
                args = {
                    keybind1 = {
                        type = "keybinding",
                        name = BINDING_NAME_ACTIONBUTTON1,
                        set = function(info, val) FA.db.global.KEYBIND1 = val end,
                        get = function(info) return FA.db.global.KEYBIND1 end,
                        order = 10,
                    },
					keybind2 = {
                        type = "keybinding",
                        name = BINDING_NAME_ACTIONBUTTON2,
                        set = function(info, val) FA.db.global.KEYBIND2 = val end,
                        get = function(info) return FA.db.global.KEYBIND2 end,
                        order = 20,
                    },
                },
            },
        },
	}

    AceConfig:RegisterOptionsTable("FlapAway", options)
    AceConfigDialog:AddToBlizOptions("FlapAway")

    FA.FAFrame.MacroButton = CreateFrame("Button", "FlapAwayButton", FA.FAFrame, "InsecureActionButtonTemplate")
    FA.FAFrame.MacroButton:SetAttribute("type1", "macro")

    FA.FAFrame.MacroButton:SetText("FlapAway")
    FA.FAFrame.MacroButton:HookScript("OnClick", FA.Update)
    FA.FAFrame:SetScript("OnKeyDown", FA.Update)
    FA.FAFrame:SetPropagateKeyboardInput(true)

    FA.FAFrame.HotkeyFontString1 = FA.FAFrame:CreateFontString("FlapAwayFrameHotkeyFontString", "OVERLAY", "GameFontWhite", 0)
    FA.FAFrame.HotkeyFontString1:SetWordWrap(false)
    FA.FAFrame.HotkeyFontString1:SetWidth(280)
    FA.FAFrame.HotkeyFontString1:SetHeight(16)
	FA.FAFrame.HotkeyFontString1:SetJustifyH("LEFT")
	
    FA.FAFrame.HotkeyFontString2 = FA.FAFrame:CreateFontString("FlapAwayFrameHotkeyFontString", "OVERLAY", "GameFontWhite", 0)
    FA.FAFrame.HotkeyFontString2:SetWordWrap(false)
    FA.FAFrame.HotkeyFontString2:SetWidth(280)
    FA.FAFrame.HotkeyFontString2:SetHeight(16)
	FA.FAFrame.HotkeyFontString2:SetJustifyH("LEFT")

	FA.FAFrame.HotkeyFontString1:SetPoint("TOPLEFT", 22, -44)
	FA.FAFrame.HotkeyFontString2:SetPoint("TOPLEFT", FA.FAFrame.HotkeyFontString1, "BOTTOMLEFT", 0, -5)

    -- Final setup
    FA:PopulateLookupTables()
	
    FA:Print(L["Addon Enabled"])
end

function FA:OnDisable()
    FA.Close()
    FA:Print(L["Addon Disabled"])
end

-- ref: https://wow.gamepedia.com/UIOBJECT_GameTooltip
-- /dump FlapAway:GetNPCName(12345)
function FA:GetNPCName(npcID)
    FA.Debug(2, "Running GetNPCName")
    FA.Debug(2, "npcID: " .. tostring(npcID))

    if npcID == nil then
        return 
    end

    if FA.npcInfo[npcID] and FA.npcInfo[npcID].npcName then
        FA.Debug(2, "NPC found in FA.npcInfo - returning result from table.  npcName: " .. tostring(FA.npcInfo[npcID].npcName))
        return FA.npcInfo[npcID].npcName
    end

    local npcName = FA:ScanFromTooltip("unit:Creature-0-0-0-0-%s-0000000000", npcID)
    if npcName == nil then
        FA.Debug(2, "npcName: not found")
    end

    FA.Debug(1, "NPC name found from tooltip - saving in FA.npcInfo.  npcName: " .. tostring(npcName))
    FA.npcInfo[npcID].npcName = npcName
    
    return npcName
end

function FA:GetQuestName(questID)
    FA.Debug(1, "Running GetQuestName")
    FA.Debug(1, "questID: " .. tostring(questID))

    if questID == nil then
        return
    end

    if FA.questName[questID] then
        FA.Debug(1, "Quest found in FA.questName - returning result from table.  questName: " .. tostring(FA.questName[questID]))
        return FA.questName[questID]
    end

    local questName = C_TaskQuest.GetQuestInfoByQuestID(questID) or C_QuestLog.GetTitleForQuestID(questID)
    -- local questName = FA:ScanFromTooltip("quest:%s", questID)
    if questName == nil then
        FA.Debug(1, "questName: not found")
    end

    FA.Debug(1, "Quest name found from tooltip - saving in FA.questName.  questName: " .. tostring(questName))
    FA.questName[questID] = questName
    
    return questName
end

--/dump FlapAway:ScanFromTooltip("quest:%s", 41944)
function FA:ScanFromTooltip(linkTemplate, id)
    if not FA.ScanningTooltip then
        FA.ScanningTooltip = CreateFrame("GameTooltip", "FlapAwayScanningTooltip", FA.FAFrame, "GameTooltipTemplate")
    end
    FA.ScanningTooltip:SetOwner(FA.FAFrame, "ANCHOR_NONE")
    FA.ScanningTooltip:SetHyperlink(format(linkTemplate, id))
    FA.Debug(1, "linkTemplate: " .. tostring(linkTemplate))
    FA.Debug(1, "id: " .. tostring(id))
    FA.Debug(1, "String: " .. format(linkTemplate, id))
    FA.Debug(1, "NumLines: " .. tostring(FlapAwayScanningTooltip:NumLines()))

    local text
    if FA.ScanningTooltip:NumLines() >= 1 then
        text = FlapAwayScanningTooltipTextLeft1:GetText()
        FA.Debug(1, "Tooltip Text: " .. text)
    else
        FA.Debug(1, "Tooltip: id not found")
    end
    return text
end

-- /dump FlapAway:PopulateLookupTables()
-- run this on initialise
function FA:PopulateLookupTables()
    FA.Debug(1, "Running PopulateLookupTables")
    local FoundAll = true
    for npcID, npcDetails in pairs(FA.npcInfo) do
        -- local questID = npcDetails.questID
        FoundAll = FoundAll and (FA:GetNPCName(npcID) ~= nil)
        if npcDetails.questID then
            FoundAll = FoundAll and (FA:GetQuestName(npcDetails.questID) ~= nil)
        end
    end
    FA.Debug(1, "FoundAll: " .. tostring(FoundAll))

    if not FoundAll then
        if FA.npcInfoTimerCount < 5 then
            FA.npcInfoTimerCount = FA.npcInfoTimerCount + 1
            FA.PopulateLookupTablesTimer = FA:ScheduleTimer("PopulateLookupTables", 5)
            FA.Debug(1, "Started timer to rerun PopulateLookupTables.  FA.npcInfoTimerCount: " .. tostring(FA.npcInfoTimerCount))
        else
            FA.Debug(1, "Max runs reached for PopulateLookupTables.  FA.npcInfoTimerCount: " .. tostring(FA.npcInfoTimerCount))
        end
    end

    FA.npcInfoFoundAll = FoundAll
    return FoundAll
end


function FA:RefreshConfig()
    FA.Debug(1, "Profile refresh detected.  Running UpdateKeybind().")
    FA.UpdateKeybind()
end

function FA.UpdateKeybind()
    if FA.db.global.KEYBIND1 == nil or FA.db.global.KEYBIND1 == "" then
        FA.FAFrame.HotkeyFontString1:SetText(BINDING_NAME_ACTIONBUTTON1 .. ": <" .. L["Keybind Missing"] .. ">")
    else
        FA.FAFrame.HotkeyFontString1:SetText(BINDING_NAME_ACTIONBUTTON1 .. ": <" .. FA.db.global.KEYBIND1 .. ">")
    end
	
    if FA.db.global.KEYBIND2 == nil or FA.db.global.KEYBIND2 == "" then
        FA.FAFrame.HotkeyFontString2:SetText(BINDING_NAME_ACTIONBUTTON2 .. ": <" .. L["Keybind Missing"] .. ">")
    else
        FA.FAFrame.HotkeyFontString2:SetText(BINDING_NAME_ACTIONBUTTON2 .. ": <" .. FA.db.global.KEYBIND2 .. ">")
    end
end

function FA.SetMacro(macroText)
    if not InCombatLockdown() then
        FA.Debug(2, "Setting macro to " .. tostring(macroText))
        FA.FAFrame.MacroButton:SetAttribute("macrotext1", macroText)
    end
end

function FA:SlashHandler(input)
    FA.Debug(1, "Running: SlashHandler")
    -- local arg1, arg2 = FA:GetArgs(input, 2, 1)
    -- FA.Debug(1, "arg1: " .. tostring(arg1) .. " arg2: " .. tostring(arg2))
    -- if arg1 and string.lower(arg1) == "config" then
    --     AceConfigDialog:Open("FlapAway")
    -- elseif arg1 and string.lower(arg1) == "reset" then
    --     FA:ResetOptions()
    -- elseif FA.FAFrame:IsShown() then
    --     FA.Close()
    -- else
    --     FA.Open()
    -- end
    local arg1, arg2 = FA:GetArgs(input, 2, 1)
    FA.Debug(1, "arg1: " .. tostring(arg1) .. " arg2: " .. tostring(arg2))

    if arg1 then
        if string.lower(arg1) == string.lower("config") then
            AceConfigDialog:Open("FlapAway")
        elseif string.lower(arg1) == "reset" then
            FA:ResetOptions()
        elseif string.lower(arg1) == "help" then
            FA:PrintHelp()
        else
            FA:Print(L["Invalid Command"])
            FA:PrintHelp()
        end
        -- HELP_LABEL
        -- RESET
        -- SETTINGS (or "config")
    elseif FA.FAFrame:IsShown() then
        FA.Close()
    else
        FA.Open()
    end
end

function FA:PrintHelp()
    FA:Print(L["Help Header"])
    FA:Print("/flapaway - " .. L["Help Default"])
    FA:Print("/flapaway config - " .. L["Help Config"])
    FA:Print("/flapaway reset - " .. L["Help Reset"])
    FA:Print("/flapaway help - " .. L["Help Help"])
end

function FA:ResetOptions()
    FA.Debug(1, "Running ResetOptions:")
    if not InCombatLockdown() then
        FA.Debug(1, "Not in combat lockdown")
        -- run reset
        FA:Print(L["Resetting Options"])
        FA.Close()
        FA.db:ResetDB("Default")
        FA.FAFrame:Reset()
        FA.Open()
        -- Refresh any open options windows
        AceConfigRegistry:NotifyChange("FlapAway")
        -- clear flag and timer
        FA.ResetOptionsAwaitingOOC = false
        FA:CancelTimer(FA.ResetOptionsOOCTimer)
    else
        if not(FA.ResetOptionsAwaitingOOC) then
            FA:Print(L["Resetting Options OOC"])
        end
        FA.ResetOptionsAwaitingOOC = true
        FA.ResetOptionsOOCTimer = FA:ScheduleTimer("ResetOptions", 1)
    end
end

function FA.Close()
    FA.Debug(1, "Close: Combat Lockdown:" .. tostring(InCombatLockdown()))
    SetCVar("AutoInteract", 0)
    if not InCombatLockdown() then
        ClearOverrideBindings(FA.FAFrame)
    end
    FA:CancelAllTimers()
    FA:ClearWaitFlag()
    -- FA:UnregisterAllEvents()
    FA:UnregisterEvent("GOSSIP_SHOW")
    FA.FAFrame:EnableKeyboard(false)
    FA.FAFrame:Hide()
    FA.Debug(1, "Frame hidden")
end

function FA.Open()
    FA.FAFrame:Show()
    FA.Debug(1, "Frame shown")
    FA:RegisterEvent("GOSSIP_SHOW", FA.EventHandler)
    FA:RegisterEvent("PLAYER_REGEN_DISABLED", FA.EventHandler)
	FA:RegisterEvent("UNIT_ENTERED_VEHICLE", FA.EventHandler)
	FA:RegisterEvent("UNIT_EXITED_VEHICLE", FA.EventHandler)
    FA.FAFrame:EnableKeyboard(true)
    FA.Update()
	FA.UpdateTimer = FA:ScheduleRepeatingTimer("Update", 1)
    FA.UpdateKeybind()
end

function FA.EventHandler(eventName)
    FA.Debug(1, "Event: " .. (eventName or ""))
    if eventName == "PLAYER_REGEN_DISABLED" then
        FA.Close()
    elseif eventName == "GOSSIP_SHOW" or eventName == "UNIT_ENTERED_VEHICLE" or eventName == "UNIT_EXITED_VEHICLE" then
		FA.Update()
	else
        FA.Debug(1, "Unknown Event: " .. (eventName or ""))
    end
end

-- in combat, do nothing
-- no keybinds, do nothing
-- 

function FA.Update()
    -- runs when registered events are triggered, after certain event, on button clicks, and on a timer
    FA.Debug(2, "Running Update")
    local hotkey1 = FA.db.global.KEYBIND1
	local hotkey2 = FA.db.global.KEYBIND2
	local InstructorNPCName = FA:GetNPCName(FA.InstructorNPCID)
	local npcID = FA:TargetNPCID()
	local ButtonName1 = "FlapAwayButton1"
	local ButtonName2 = "FlapAwayButton2"
	
	-- if ElvUI_Bar1 then
		-- FA.Debug(2, "ElvUI_Bar1 found - using ElvUI buttons.")
		-- ButtonName1 = "ElvUI_Bar1Button1"
		-- ButtonName2 = "ElvUI_Bar1Button2"
	-- else
		-- FA.Debug(2, "ElvUI_Bar1 not found - using default UI buttons.")
		-- ButtonName1 = "OverrideActionBarButton1"
		-- ButtonName2 = "OverrideActionBarButton2"
	-- end

    FA.UpdateKeybind()

    if InCombatLockdown() then
        FA.Debug(2, "In combat - take no action")
    elseif hotkey1 == nil or hotkey2 == nil then
        FA.Debug(2, "Hotkeys not set - take no action")
	elseif UnitInVehicle("player") and (IsSpellKnown(332710, true) or IsSpellKnown(342976, true)) then
		FA.Debug(2, "Player in vehicle, and Flap or Begin spells known.  Verify keybinds set.")
		SetCVar("AutoInteract", 0)
		-- ClearOverrideBindings(FA.FAFrame)
		-- /dump GetBindingAction("MOUSEWHEELDOWN", true)
		-- /dump GetBindingAction("MOUSEWHEELDOWN", true):find("CLICK OverrideActionBarButton1")
		if not (GetBindingAction(hotkey1, true):find("CLICK " .. ButtonName1) == 1 and GetBindingAction(hotkey2, true):find("CLICK " .. ButtonName2) == 1) then
			FA.Debug(2, "OverrideActionButtons keybinds not set - applying now.")
			ClearOverrideBindings(FA.FAFrame)
			SetOverrideBindingClick(FA.FAFrame, true, hotkey1, ButtonName1)
			SetOverrideBindingClick(FA.FAFrame, true, hotkey2, ButtonName2)
		else
			FA.Debug(2, "Keybinds confirmed.")
		end
    elseif GossipFrame:IsShown() and GossipFrameNpcNameText:GetText() == InstructorNPCName then
		FA.Debug(2, "Gossip window open and NPC matches.  Select gossip option.")
		SetCVar("AutoInteract", 0)
		ClearOverrideBindings(FA.FAFrame)
		if C_QuestLog.IsQuestFlaggedCompleted(60911) or C_GossipInfo.GetNumOptions() == 1 or IsShiftKeyDown() then
			C_GossipInfo.SelectOption(1, "", true)
		else
			C_GossipInfo.SelectOption(3, "", true)
		end
	elseif (npcID == FA.InstructorNPCID) then
		FA.Debug(2, "Target matches.  Set button to IWT and enable CTM.")
		SetCVar("AutoInteract", 1)
		ClearOverrideBindings(FA.FAFrame)
		SetOverrideBinding(FA.FAFrame, true, hotkey1, "INTERACTTARGET")
		SetOverrideBinding(FA.FAFrame, true, hotkey2, "INTERACTTARGET")
	else
		FA.Debug(2, "Set macro to /targetexact [NPC].  Disable IWT and CTM.")
		SetCVar("AutoInteract", 0)
		ClearOverrideBindings(FA.FAFrame)
		FA.SetMacro("/targetexact " .. InstructorNPCName)
		SetOverrideBindingClick(FA.FAFrame, true, hotkey1, "FlapAwayButton")
		SetOverrideBindingClick(FA.FAFrame, true, hotkey2, "FlapAwayButton")
	end
end

function FA:TargetNPCID()
    if UnitGUID("target") then
        return tonumber((select(6, strsplit("-", UnitGUID("target")))))
    end
end

function FA:WaitForTimer(seconds)
    -- Pause any actions for a set period
    local TimeRemaining = FA:TimeLeft(FA.WaitTimer)
    if TimeRemaining > seconds then
        FA.Debug(1, "WaitFlag timer already running, with " .. tostring(TimeRemaining) .. "s left.  Already covers new timer, so no action taken.")
        return
    elseif TimeRemaining > 0 then
        FA.Debug(1, "WaitFlag timer already running, with " .. tostring(TimeRemaining) .. "s left.  Resetting to start new run.")
        FA:CancelTimer(FA.WaitTimer)
    end
    FA.WaitFlag = true
    FA.WaitTimer = FA:ScheduleTimer(FA.ClearWaitFlag, seconds)
    FA.Debug(1, "WaitFlag set.  Time started for " .. tostring(seconds) .. "s.")
end

function FA.ClearWaitFlag()
    -- Clear Wait Flag - only to be called by FA:WaitForTimer
    if FA.WaitFlag then
        FA.Debug(1, "Clearing WaitFlag.")
    else
        FA.Debug(1, "WaitFlag already cleared.")
    end
    FA.WaitFlag = false
end

function FA:PLAYER_REGEN_DISABLED(eventName)
    FA.EventHandler(eventName)
end